

#include "HexDisplay.h"
#include <math.h>
#include "const7.h"

#include <iostream>
#include <stdlib.h>

#include <SDL2/SDL2_gfxPrimitives.h>

using namespace std;

HexDisplay::HexDisplay() : LineColor(22, 22, 22), FillColor(33, 33, 33) {
	// initialize default radius?  really it should be set though
	Radius = 1;

	FillEnabled = false;
}


void HexDisplay::setRadius(double r) {
	Radius = r;

	// calculate once to avoid wasting cpu, and we only need the rounded int values anyway
	Half_R_Sqrt3 = Radius * SQRT_3 / 2;
	Half_R = Radius / 2;
}

void  HexDisplay::DrawHex(SDL_Renderer* renderer, int location_x, int location_y) {
	// draw 6 lines for the sides

	// calculate hex points
	short int xpos[6], ypos[6];

	// need to avoid integer overflows for hexes positioned far, far off the screen
	// also probably wouldn't hurt to clip for efficiency too, if it can be done efficiently...

	// much wider than any screen window we are likely to have
	if (location_x - Radius > 5000 ||
		location_y - Radius > 5000 ||
		location_x + Radius < 0 ||
		location_y + Radius < 0
		) {
		  //  cout << "ignoring out of bounds location: " << location_x << "," << location_y << endl;

		return;
	}


	xpos[0] = location_x + Radius;
	ypos[0] = location_y;
	xpos[1] = location_x + Half_R;
	ypos[1] = location_y - Half_R_Sqrt3;
	xpos[2] = location_x - Half_R;
	ypos[2] = location_y - Half_R_Sqrt3;
	xpos[3] = location_x - Radius;
	ypos[3] = location_y;
	xpos[4] = location_x - Half_R;
	ypos[4] = location_y + Half_R_Sqrt3;
	xpos[5] = location_x + Half_R;
	ypos[5] = location_y + Half_R_Sqrt3;


	 SDL_SetRenderDrawColor(renderer, LineColor.getred(), LineColor.getgreen(), LineColor.getblue(), LineColor.getAlpha());

	int nexti;




	for (int i = 0; i < 6; i++) {
		nexti = i+1;
		if (nexti > 5) nexti = 0;
		SDL_RenderDrawLine(renderer,
					   xpos[i],
					   ypos[i],


					   xpos[nexti],
					   ypos[nexti]
					   );
	}

	if (FillEnabled) {
		// draw filled poly

		xpos[0]--;
		xpos[1]--;
		ypos[1]++;
		xpos[2]++;
		ypos[2]++;
		xpos[3]++;
		xpos[4]++;
		ypos[4]--;
		xpos[5]--;
		ypos[5]--;


					 filledPolygonRGBA(renderer,
					   xpos,
					   ypos,
					   6,
					   FillColor.getred(),
					   FillColor.getgreen(),
					   FillColor.getblue(),
					   FillColor.getAlpha()
					   );

	}


}

void HexDisplay::DrawInvertedHex(SDL_Renderer* renderer, int location_x, int location_y ) {
		// draw 6 lines for the sides

	// calculate hex points
	short int xpos[6], ypos[6];

	// need to avoid integer overflows for hexes positioned far, far off the screen
	// also probably wouldn't hurt to clip for efficiency too, if it can be done efficiently...

	// much wider than any screen window we are likely to have
	if (location_x - Radius > 5000 ||
		location_y - Radius > 5000 ||
		location_x + Radius < 0 ||
		location_y + Radius < 0
		) {
		   // cout << "ignoring out of bounds location: " << location_x << "," << location_y << endl;
		return;
	}

	xpos[0] = location_x + Half_R_Sqrt3;
	ypos[0] = location_y - Half_R;
	xpos[1] = location_x;
	ypos[1] = location_y - Radius;
	xpos[2] = location_x - Half_R_Sqrt3;
	ypos[2] = location_y - Half_R;
	xpos[3] = location_x - Half_R_Sqrt3;
	ypos[3] = location_y + Half_R;
	xpos[4] = location_x;
	ypos[4] = location_y + Radius;
	xpos[5] = location_x + Half_R_Sqrt3;
	ypos[5] = location_y + Half_R;

	SDL_SetRenderDrawColor(renderer, LineColor.getred(), LineColor.getgreen(), LineColor.getblue(), SDL_ALPHA_OPAQUE);

	int nexti;

	for (int i = 0; i < 6; i++) {
		nexti = i+1;
		if (nexti > 5) nexti = 0;
		SDL_RenderDrawLine(renderer,
					   xpos[i],
					   ypos[i],
					   xpos[nexti],
					   ypos[nexti]
					   );

	}

	if (FillEnabled) {
		// draw filled poly


		// shrink by 1 pixel for fill poly
		xpos[0]--;
		ypos[0]++;
		ypos[1]++;
		xpos[2]++;
		ypos[2]++;
		xpos[3]--;
		ypos[3]--;
		ypos[4]--;
		xpos[5]--;
		ypos[5]--;


			 filledPolygonRGBA(renderer,
					   xpos,
					   ypos,
					   6,
					   FillColor.getred(),
					   FillColor.getgreen(),
					   FillColor.getblue(),
					   FillColor.getAlpha()
					   );


	}

}



