#include "HexSpiralCoordinate.h"

#include <math.h>
#include <iostream>

using namespace std;

HexSpiralCoordinate::HexSpiralCoordinate(long int hex_index) {
	HexIndex = hex_index;
	calcLocationInfo();
}


// based on a given index in the spiral, calculates:
//  * rdistance : the radial distance (in hexagon step units) from the center
//  * circ : circumference of the hex-grid at that point in the spiral
//  * phase : which "side" of the larger hex shape this one is on (0-5)
//  * poffset: how many hexes from the beginning of the phase it is

void HexSpiralCoordinate::calcLocationInfo() {

	//HexSpiralCoordinate info;

	rdistance = 0;
	circ = 1;  // current circumference... increases by 6 with each round (initial value of 1 = single center hex)

	long int spiralidx = 0;

	if (HexIndex > 0) {

		// summation, but inverted:
		//int m = HexIndex / 6;
		//int sum = 6 * (m * (m + 1) / 2);
		//int ans = 6 * sum;

		// r^2 + r = HexIndex/3

		/*
		do  {
			spiralidx += circ;
			rdistance++;

			circ = rdistance * 6;
		} while (spiralidx + circ <= HexIndex);
		*/

		// is this really more efficient than the above loop?  I'm really not sure
		rdistance = ceil( (sqrt(12*HexIndex+9) - 3) / 6 );
		circ = rdistance * 6;
		spiralidx = 1 + 6 * ((rdistance - 1) * (rdistance) / 2 );


		phase = (HexIndex - spiralidx ) / rdistance;
		poffset = (HexIndex - spiralidx ) % rdistance;
	} else {
		phase = 0;
		poffset = 0;
	}

	// cout << "info idx: " << HexIndex << "  rdistance: " << rdistance << "  circ: " << circ << "  spiralidx: " << spiralidx << "  phase: "  << phase << "  poffset: " << poffset << endl;
}

HexSpiralCoordinate::HexSpiralCoordinate(long int hex_index, int p_rdistance, int p_phase, int p_poffset, int p_circ) {
	HexIndex = hex_index;
	rdistance = p_rdistance;
	phase = p_phase;
	poffset = p_poffset;
	circ = p_circ;
}


HexSpiralCoordinate::HexSpiralCoordinate(int i_rdistance, int i_phase, int i_poffset) {
	rdistance = i_rdistance;
	phase = i_phase;
	poffset = i_poffset;

	// calculate circ and HexIndex from above vals
	circ = rdistance * 6;
	if (rdistance > 0) {
		//HexIndex = 1 + (rdistance - 1) * 6 + (phase * rdistance) + poffset;
		HexIndex = 1 + 6 * ((rdistance - 1) * (rdistance) / 2 ) + (phase * rdistance) + poffset;
	} else {
		HexIndex = 0;
	}

}


// calculates the hex index in a spiral a unit away in given direction (0-5) from oldidx
HexSpiralCoordinate HexSpiralCoordinate::calcMovement(  //int oldidx, HexSpiralCoordinate& hexinfo,
											  int direction) {

	// will contain the new spiral index after movement has been added
	HexSpiralCoordinate newhex = *this;

	if (HexIndex > 0) {
		/*
		cout << "HexIndex: " << HexIndex << "  direction: " << direction
			<< "  rdistance: " << rdistance
			<< "  circ: " << circ
			<< "  phase: "  << phase
			<< "  poffset: " << poffset
			<< endl;
			*/

		// based on which direction we want to go,
		// and the current position "phase" (which of the six sides of the larger hexgrid it is on),
		// the calculation of each step

		switch (direction) {
			case 0: // semi-northeast / east
				switch (phase) {
					case 0:
						newhex.poffset++;
					case 1:
						newhex.HexIndex = HexIndex + newhex.circ + 1;
						newhex.rdistance++;
						newhex.circ+=6;
						break;
					case 2:
						newhex.HexIndex = HexIndex - 1;
						if (newhex.poffset == 0) {
							newhex.phase--;
							newhex.poffset = newhex.rdistance - 1;
						} else {
							newhex.poffset--;
						}
						break;
					case 3:
						newhex.HexIndex = HexIndex - newhex.circ + 2;
						if (newhex.rdistance > 1) {
							newhex.rdistance--;
							newhex.circ-=6;
							if (newhex.poffset == 0) {
								newhex.phase--;
								newhex.poffset = newhex.rdistance - 1;
							} else {
								newhex.poffset--;
							}
						} else {
							newhex.rdistance = 0;
							newhex.phase = 0;
							newhex.poffset = 0;
							newhex.circ = 1;
						}
						break;
					case 4:
						if (newhex.poffset == newhex.rdistance - 1) {
							newhex.HexIndex = HexIndex + 1;
							newhex.phase++;
							newhex.poffset=0;
						} else {
							newhex.HexIndex = HexIndex - newhex.circ + 2;
							newhex.rdistance--;
							newhex.circ-=6;
						}
						break;

					case 5:
						newhex.HexIndex = HexIndex + 1;
						if (newhex.poffset == newhex.rdistance - 1) {
							newhex.rdistance++;
							newhex.phase = 0;
							newhex.poffset = 0;
							newhex.circ += 6;
						} else {
							newhex.poffset++;
						}
						break;
				};
				break;

			case 1: // north / northeast
				switch (newhex.phase) {
					case 0:
						if (newhex.poffset == newhex.rdistance - 1) {
							newhex.HexIndex = HexIndex + newhex.circ + 2;
							newhex.rdistance++;
							newhex.phase++;
							newhex.poffset=0;
							newhex.circ+=6;
						} else {
							newhex.HexIndex = HexIndex + 1;
							newhex.poffset++;
						}
						break;
					case 1:
						newhex.poffset++;
					case 2:
						newhex.HexIndex = HexIndex + newhex.circ + 2;
						newhex.rdistance++;
						newhex.circ+=6;
						break;
					case 3:
						newhex.HexIndex = HexIndex - 1;

						if (newhex.poffset > 0) {
							newhex.poffset--;
						} else {
							newhex.poffset = newhex.rdistance - 1;
							newhex.phase--;
						}

						break;
					case 4:
						newhex.HexIndex = HexIndex - newhex.circ + 1;
						if (newhex.rdistance > 1) {
							newhex.rdistance--;
							newhex.circ-=6;
							if (newhex.poffset > 0) {
								newhex.poffset--;
							} else {
								newhex.phase--;
								newhex.poffset = newhex.rdistance - 1;
							}
						} else {
							newhex.rdistance = 0;
							newhex.poffset = 0;
							newhex.phase = 0;
							newhex.circ = 1;
						}
						break;
					case 5:
						newhex.HexIndex = HexIndex - newhex.circ + 1;
						if (newhex.poffset == newhex.rdistance - 1) {
							newhex.phase = 0;
							newhex.poffset = 0;
						} else {
							newhex.rdistance--;
							newhex.circ-=6;
						}
						break;
				}
				break;

			case 2: // semi-northwest / northwest
				switch (newhex.phase) {
					case 0:
						if (newhex.poffset == newhex.rdistance - 1) {
							newhex.HexIndex = HexIndex + 1;
							newhex.phase++;
							newhex.poffset = 0;
						} else {
							newhex.HexIndex = HexIndex - (newhex.circ - 6);
							newhex.rdistance--;
							newhex.circ-=6;
						}
						break;
					case 1:
						if (newhex.poffset == newhex.rdistance - 1) {
							newhex.HexIndex = HexIndex + newhex.circ + 3;
							newhex.rdistance++;
							newhex.phase++;
							newhex.poffset = 0;
							newhex.circ+=6;
						} else {
							newhex.HexIndex = HexIndex + 1;
							newhex.poffset++;
						}
						break;
					case 2:
						newhex.poffset++;
					case 3:
						newhex.HexIndex = HexIndex + newhex.circ + 3;
						newhex.rdistance++;
						newhex.circ+=6;
						break;
					case 4:
						newhex.HexIndex = HexIndex - 1;
						if (newhex.poffset > 0)
							newhex.poffset--;
						else {
							newhex.poffset = newhex.rdistance - 1;
							newhex.phase--;
						}
						break;
					case 5:
						newhex.HexIndex = HexIndex - newhex.circ;
						newhex.rdistance--;
						newhex.circ-=6;
						if (newhex.rdistance > 0) {
							if (newhex.poffset > 0) {
								newhex.poffset--;
							} else {
								newhex.phase--;
								newhex.poffset = newhex.rdistance - 1;
							}
						} else {
							newhex.phase = 0;
							newhex.poffset = 0;
						}
						break;
				};
				break;

			case 3: // semi-southwest / west
				switch (newhex.phase) {
					case 0:
						if (newhex.rdistance > 1) {
							if (newhex.poffset == 0) {
								newhex.HexIndex = HexIndex - 1;
								newhex.rdistance--;
								newhex.circ -= 6;
								newhex.phase = 5;
								newhex.poffset = newhex.rdistance - 1;
							} else {
								newhex.HexIndex = HexIndex - (newhex.circ - 6) - 1;
								newhex.rdistance--;
								newhex.circ -= 6;
								newhex.poffset--;
							}
						} else {
							newhex.HexIndex = 0;
							newhex.rdistance = 0;
							newhex.phase = 0;
							newhex.poffset = 0;
							newhex.circ = 1;
						}
						break;
					case 1:
						if (newhex.poffset == newhex.rdistance - 1) {
							newhex.HexIndex = HexIndex + 1;
							newhex.poffset = 0;
							newhex.phase++;
						} else {
							newhex.HexIndex = HexIndex - (newhex.circ - 6) - 1;
							newhex.rdistance--;
							newhex.circ-=6;
						}
						break;
					case 2:
						if (newhex.poffset == newhex.rdistance - 1) {
							newhex.HexIndex = HexIndex + newhex.circ + 4;
							newhex.rdistance++;
							newhex.circ+=6;
							newhex.phase++;
							newhex.poffset=0;
						} else {
							newhex.HexIndex = HexIndex + 1;
							newhex.poffset++;
						}
						break;
					case 3:
						newhex.poffset++;
					case 4:
						newhex.HexIndex = HexIndex + newhex.circ + 4;
						newhex.rdistance++;
						newhex.circ+=6;
						break;
					case 5:
						newhex.HexIndex = HexIndex - 1;
						if (newhex.poffset == 0) {
							newhex.phase--;
							newhex.poffset = newhex.rdistance - 1;
						} else
							newhex.poffset--;
						break;
				};
				break;

			case 4: // south / southwest
				switch (newhex.phase) {
					case 0:
						if (newhex.poffset == 0) {
							newhex.HexIndex = HexIndex + newhex.circ - 1;
							newhex.phase = 5;
							newhex.poffset = newhex.rdistance - 1;
						} else {
							newhex.HexIndex = HexIndex - 1;
							newhex.poffset--;
						}
						break;
					case 1:
						newhex.HexIndex = HexIndex - (newhex.circ - 6) - 2;
						if (newhex.rdistance > 1) {
							newhex.rdistance--;
							newhex.circ-=6;
							if (newhex.poffset == 0) {
								newhex.phase--;
								newhex.poffset = newhex.rdistance - 1;
							} else {
								newhex.poffset--;
							}
						} else {
							newhex.rdistance = 0;
							newhex.phase = 0;
							newhex.poffset = 0;
							newhex.circ=1;
						}
						break;
					case 2:
						if (newhex.poffset == newhex.rdistance - 1) {
							newhex.HexIndex = HexIndex + 1;
							newhex.phase++;
							newhex.poffset=0;
						} else {
							newhex.HexIndex = HexIndex - (newhex.circ - 6) - 2;
							newhex.rdistance--;
							newhex.circ-=6;
						}
						break;
					case 3:
						if (newhex.poffset == newhex.rdistance - 1) {
							newhex.HexIndex = HexIndex + newhex.circ + 5;
							newhex.rdistance++;
							newhex.circ+=6;
							newhex.phase++;
							newhex.poffset=0;
						} else {
							newhex.HexIndex = HexIndex + 1;
							newhex.poffset++;
						}
						break;
					case 4:
						newhex.poffset++;
					case 5:
						newhex.HexIndex = HexIndex + newhex.circ + 5;
						newhex.rdistance++;
						newhex.circ+=6;
						break;
				}
				break;

			case 5: // semi-southeast / southeast
				switch (newhex.phase) {
					case 0:
						newhex.HexIndex = HexIndex + newhex.circ;
						newhex.rdistance++;
						newhex.circ+=6;
						break;
					case 1:
						newhex.HexIndex = HexIndex - 1;
						if (newhex.poffset == 0) {
							newhex.phase--;
							newhex.poffset = newhex.rdistance - 1;
						} else {
							newhex.poffset--;
						}
						break;
					case 2:
						newhex.HexIndex = HexIndex - (newhex.circ - 6) - 3;
						if (newhex.rdistance > 1) {
							newhex.rdistance--;
							newhex.circ-=6;
							if (newhex.poffset == 0) {
								newhex.phase--;
								newhex.poffset = newhex.rdistance - 1;
							} else {
								newhex.poffset--;
							}
						} else {
							newhex.phase = 0;
							newhex.poffset = 0;
							newhex.rdistance = 0;
							newhex.circ = 1;
						}
						break;
					case 3:
						if (newhex.poffset == newhex.rdistance - 1) {
							newhex.HexIndex = HexIndex + 1;
							newhex.phase++;
							newhex.poffset=0;
						} else {
							newhex.HexIndex = HexIndex - (newhex.circ - 6) - 3;
							newhex.rdistance--;
							newhex.circ-=6;
						}
						break;
					case 4:
						if (newhex.poffset == newhex.rdistance - 1) {
							newhex.HexIndex = HexIndex + newhex.circ + 6;
							newhex.rdistance++;
							newhex.circ+=6;
							newhex.phase++;
							newhex.poffset=0;
						} else {
							newhex.HexIndex = HexIndex + 1;
							newhex.poffset++;
						}
						break;
					case 5:
						newhex.HexIndex = HexIndex + newhex.circ + 6;
						newhex.rdistance++;
						newhex.circ+=6;
						newhex.poffset++;
						break;
				}
		}
	} else { // HexIndex == 0, dead center, special case

		// simple case:  moving out from center
		newhex.HexIndex = direction + 1;

		newhex.rdistance = 1;
		newhex.phase = direction;
		newhex.poffset = 0;
		newhex.circ = 6;

	}

	/*
	cout << "newhex.HexIndex: " << newhex.HexIndex << "  direction: " << direction
			<< "  rdistance: " << newhex.rdistance
			<< "  circ: " << newhex.circ
			<< "  phase: "  << newhex.phase
			<< "  poffset: " << newhex.poffset
			<< endl << endl;
			*/

	return newhex;
}


XYLocation HexSpiralCoordinate::getXY(bool is_inverted)
{
	XYLocation hexxy;

	if (rdistance == 0) {
		hexxy.setx(0);
		hexxy.sety(0);
	} else if (!is_inverted) {
		switch (phase) {
			case 0:
				hexxy.setx(rdistance);
				hexxy.sety(  -1 + (rdistance - 1) - (poffset * 2) );
				break;
			case 1:
				hexxy.setx((rdistance - 1) - poffset);
				hexxy.sety( - (1 + rdistance + poffset )  );
				break;
			case 2:
				hexxy.setx( (-1 - poffset) );
				hexxy.sety( -1 - ( (rdistance-1) * 2) + poffset);
				break;
			case 3:
				hexxy.setx(-rdistance);
				hexxy.sety(1 - (rdistance - 1) + (poffset * 2) );
				break;
			case 4:
				hexxy.setx( - (rdistance - 1)  + poffset );
				hexxy.sety( 1 + rdistance + poffset);
				break;
			case 5:
				hexxy.setx( 1 + poffset);
				hexxy.sety(1 + (rdistance-1) * 2 - poffset);
				break;
		}
	} else {
		switch (phase) {
			case 0:
				hexxy.setx(1 + rdistance  + poffset);
				hexxy.sety(rdistance - 1 - poffset);
				break;
			case 1:
				hexxy.setx(1 + (rdistance-1) * 2 - poffset);
				hexxy.sety(-1 - poffset);
				break;
			case 2:
				hexxy.setx(-1 + (rdistance-1) - poffset * 2);
				hexxy.sety(-rdistance);
				break;
			case 3:
				hexxy.setx(-1 - (rdistance) - poffset);
				hexxy.sety(- (rdistance - 1) + poffset);
				break;
			case 4:
				hexxy.setx(-1 - ((rdistance-1) * 2) + poffset);
				hexxy.sety(1 + poffset);
				break;
			case 5:
				hexxy.setx(1 - (rdistance - 1) + poffset * 2);
				hexxy.sety(1 + rdistance - 1);
				break;
		}
	}

   // cout << "calculated xy... x:" << hexxy.getx() << "  y:" << hexxy.gety() << endl;

	return hexxy;
}

HexSpiralCoordinate HexSpiralCoordinate::operator + (const HexSpiralCoordinate& h) const {
	int n_rdistance, n_poffset, n_phase;

	if (rdistance == 0) {
		return h;
	}
	int phasediff = h.phase - phase;

	if (phasediff < -2)
		phasediff += 6;
	else if (phasediff > 3)
		phasediff -= 6;

	n_phase = phase;

	if (h.rdistance > 0) {
	switch (phasediff) {
	case 0: // same phase
		n_rdistance = rdistance + h.rdistance;
		n_poffset = poffset + h.poffset + 1;
		break;
	case 1:
		n_rdistance = rdistance + (h.rdistance - 1 - h.poffset);
		n_poffset = poffset + h.rdistance;
		if (n_poffset >= n_rdistance) {
			n_phase = phase < 5 ? phase + 1 : 0;
			n_poffset -= n_rdistance;
			n_rdistance += n_poffset + 1;
		}
		break;
	case 2:
		n_rdistance = rdistance - 1 - h.poffset;
		n_poffset = poffset - h.poffset + (h.rdistance - 1);
		if (n_rdistance - n_poffset <= 0) {
			n_phase = phase < 5 ? phase + 1 : 0;
			int tmp_rdistance = n_rdistance;
			n_rdistance = n_poffset + 1;
			n_poffset = n_poffset - tmp_rdistance;
		}
		if (n_poffset >= n_rdistance) {
			n_phase = n_phase < 5 ? n_phase + 1 : 0;
			n_poffset -= n_rdistance;
			n_rdistance += n_poffset + 1;
		}

		break;
	case 3:
		n_rdistance = rdistance - h.rdistance;
		n_poffset = poffset - h.poffset;
		if (n_rdistance == 0) { // on a line through the center
			if (n_poffset == 0) {
				// result: origin point
				n_poffset = 0;
				n_phase = 0;
			} else if (n_poffset > 0) {
				n_phase = phase + 1 < 6 ? phase + 1 : 0;
				n_rdistance = n_poffset;
				n_poffset--;
			} else { // (n_poffset < 0)
				n_phase = n_phase - 2 >= 0 ? n_phase - 2 : n_phase + 4;
				n_rdistance = (-n_poffset);
				n_poffset = n_rdistance - 1;
			}
		} else if (n_rdistance > 0) {
			if (n_poffset > n_rdistance) {
				n_phase = phase + 1 < 6 ? phase + 1 : 0;
				int tmp_rdistance = n_rdistance;
				n_rdistance = n_poffset;
				n_poffset = n_poffset - tmp_rdistance - 1;
			} else if (n_poffset - 1 < 0) {
				n_phase = phase > 0 ? phase - 1 : 5;
				int tmp_rdistance = n_rdistance;
				n_rdistance = (-n_poffset) + n_rdistance;
				n_poffset = tmp_rdistance - 1;
			} else {
				n_poffset--;
			}
		} else { // n_rdistance < 0
			if (n_poffset >= 0) {
				n_phase = n_phase + 2 < 6 ? n_phase + 2 : n_phase - 4;
				int tmp_rdistance = n_rdistance;
				n_rdistance = (-n_rdistance) + n_poffset;
				n_poffset = (-tmp_rdistance) - 1;
			} else if (n_poffset >= n_rdistance) {
				n_phase = n_phase + 3 < 6 ? n_phase + 3 : n_phase - 3;
				n_rdistance = (-n_rdistance);
				n_poffset = (-n_poffset) - 1;
			} else {
				n_phase = n_phase + 4 < 6 ? n_phase + 4 : n_phase - 2;
				n_poffset = n_rdistance - n_poffset - 1;
				n_rdistance = (-n_rdistance) + 1 + n_poffset;
			}
		}
		break;

	case -1:
		n_rdistance = rdistance + 1 + h.poffset;
		n_poffset = poffset - (h.rdistance - 1 - h.poffset);
		if (n_poffset < 0) {
			n_phase = n_phase > 0 ? n_phase - 1 : 5;
			int tmp_rdistance = n_rdistance;
			n_rdistance = n_rdistance - 1 + (-n_poffset);
			n_poffset = tmp_rdistance - 1;
		}
		break;

	case -2:
		n_rdistance = rdistance - (h.rdistance - 1 - h.poffset);
		n_poffset = poffset - h.rdistance;
		if (n_rdistance > 0 && n_poffset < 0) {
			n_phase = n_phase > 0 ? n_phase - 1 : 5;
			int tmp_rdistance = n_rdistance;
			n_rdistance = n_rdistance - 1 + (-n_poffset);
			n_poffset = tmp_rdistance - 1;
		} else if ( n_rdistance < 1 ) {
			n_phase = n_phase - 2 >= 0 ? n_phase - 2 : n_phase + 4;
			n_poffset = (-n_poffset) - 2 - (-n_rdistance);
			n_rdistance = 1 + (-n_rdistance) + n_poffset;
		}
	}
	} else {
		n_rdistance = rdistance;
		n_poffset = poffset;
		n_phase = phase;
	}

	return HexSpiralCoordinate(n_rdistance, n_phase, n_poffset);
}

HexSpiralCoordinate HexSpiralCoordinate::operator - (const HexSpiralCoordinate& h) const {
	if (h.rdistance == 0) return (*this);

	HexSpiralCoordinate negated(h.rdistance, h.phase < 3 ? h.phase + 3 : h.phase - 3, h.poffset);
	return (*this) + negated;
}


HexSpiralCoordinate HexSpiralCoordinate::operator / (const long int &i) const {
	HexSpiralCoordinate result(rdistance / i, phase, poffset / i);
	return result;
}

HexSpiralCoordinate HexSpiralCoordinate::avg(HexSpiralCoordinate* CoordinateArray, int length) {
	HexSpiralCoordinate sum(0);
	for (int i = 0; i < length; i++) {
		sum = sum + CoordinateArray[i];
	}
	return sum / length;
}



HexSpiralCoordinate HexSpiralCoordinate::getTranslatedCoordinate(HexSpiralCoordinate offsetCoordinate)
{
	return (*this) + offsetCoordinate;
	// == old (less efficinet) algo: ==


	// calculate steps from origin (0) to offsetCoordinate
	// as a count of steps in each of the needed directions (straight_steps vs bent_steps)
	// which will always begin in the direction of offsetCoordinate.phase

	// then add steps to copied base values of *this coordinate to arrive at the sum and return result coordinate


	/*
	int initial_direction = offsetCoordinate.phase;
	int bent_direction = initial_direction > 0 ? initial_direction - 1 : 5;
	int total_steps = offsetCoordinate.rdistance;

	int bent_steps = offsetCoordinate.rdistance > 0 ? offsetCoordinate.rdistance - 1 - offsetCoordinate.poffset : 0;

	int straight_steps = total_steps - bent_steps;

	HexSpiralCoordinate nextCoord = *this;

	// initial test implementation, might come up with a less loop-intensive O(1) version later, possibly...
	int step;
	for (step = 0; step < straight_steps; step++) {
		nextCoord = nextCoord.calcMovement(initial_direction);
	}

	for (step = 0; step < bent_steps; step++) {
		nextCoord = nextCoord.calcMovement(bent_direction);
	}

	return nextCoord
	*/
}

HexSpiralCoordinate HexSpiralCoordinate::getParentSpiralCoordinate(bool direction_forward, int map_radius)
{
	int rmult = rdistance / map_radius;
	int roff = rdistance % map_radius;

	int p_phase = 0, p_rdistance = 0, p_poffset = 0;

	if (rmult >= 1) {
		p_rdistance = 1;
		p_phase = phase;

		int rstretch = rmult % 3;
		int rstretch_count = rmult / 3;

		int poffmult = poffset / map_radius;
		int poffoff = poffset % map_radius;

		int poff_stretch = poffmult % 3;
		int poff_stretch_count = poffmult / 3;


		if (direction_forward) {
		switch (rstretch) {
		case 0:
			if (
				( rstretch_count % 2 == 0 &&
					poff_stretch_count <= rstretch_count / 2
				) ||
				( rstretch_count % 2 == 1 &&
					(
						(poff_stretch_count < rstretch_count / 2) ||
						(poff_stretch_count == rstretch_count / 2 &&
							(poff_stretch < 2 || poffoff < roff)
						)
					)
				)
			)
			{
				if (poff_stretch == 0)
				{
					p_rdistance = rstretch_count * 2 - poff_stretch_count;
					p_poffset = rstretch_count - 1 + poff_stretch_count;
				}
				else if (poff_stretch == 1 || (poff_stretch == 2 && poffoff < roff) )
				{
					p_rdistance = rstretch_count * 2 - poff_stretch_count;
					p_poffset = rstretch_count + poff_stretch_count;
				} else {
					p_rdistance = rstretch_count * 2 - poff_stretch_count - 1;
					p_poffset = rstretch_count + poff_stretch_count;
				}

			} else {
				p_phase = phase < 5 ? phase + 1 : 0;

				switch (poff_stretch) {
				case 0:
					p_rdistance = rstretch_count * 2 - (rstretch_count - poff_stretch_count ) ;
					p_poffset = rstretch_count - ( ( rstretch_count - poff_stretch_count ) * 2 ) - 1;
					break;
				case 1:
					p_rdistance = rstretch_count * 2 - (rstretch_count - poff_stretch_count - 1);
					p_poffset = rstretch_count - ( ( (rstretch_count - poff_stretch_count) - 1 ) * 2 ) - 2;
				case 2:
					if (poffoff < roff) {
						p_rdistance = rstretch_count * 2 - (rstretch_count - poff_stretch_count - 1);
						p_poffset = rstretch_count - ( ( (rstretch_count - poff_stretch_count) - 1 ) * 2 ) - 2;
					} else {
						p_rdistance = rstretch_count * 2 - (rstretch_count - poff_stretch_count - 1 ) ;
						p_poffset = rstretch_count - ( ( rstretch_count - poff_stretch_count - 1) * 2 ) - 1;
					}
				}
			}
			break;

		case 1:
			if (
				(rstretch_count % 2 == 1 &&
					poff_stretch_count <= rstretch_count / 2
				) ||
				(rstretch_count % 2 == 0 && (
					(poff_stretch_count < rstretch_count / 2 ) ||
					( (poff_stretch_count == rstretch_count / 2) &&
						(poff_stretch == 0 ||
							(poff_stretch == 1 && poffoff < roff)
						)
					)
				) )
			)
			{
				switch (poff_stretch) {
				case 0:
					p_rdistance = rstretch_count * 2 + 1 - poff_stretch_count;
					p_poffset = rstretch_count + poff_stretch_count;
					break;
				case 1:
					if (poffoff < roff) {
						p_rdistance = rstretch_count * 2 + 1 - poff_stretch_count;
						p_poffset = rstretch_count + poff_stretch_count;
					} else {
						p_rdistance = rstretch_count * 2 - poff_stretch_count;
						p_poffset = rstretch_count + poff_stretch_count;
					}
					break;
				case 2:
					p_rdistance = rstretch_count * 2 - poff_stretch_count;
					p_poffset = rstretch_count + poff_stretch_count;
					break;
				}
			}
			else
			{
				p_phase = phase < 5 ? phase + 1 : 0;

				switch (poff_stretch) {
				case 0:
					p_rdistance = rstretch_count * 2 + 1 - (rstretch_count - poff_stretch_count);
					p_poffset = rstretch_count - 1 - (rstretch_count - poff_stretch_count) * 2;
					break;
				case 1:
					if (poffoff < roff) {
						p_rdistance = rstretch_count * 2 + 1 - (rstretch_count - poff_stretch_count);
						p_poffset = rstretch_count - 1 - (rstretch_count - poff_stretch_count) * 2;
					} else {
						if (rstretch_count % 2 == 0) {
							p_rdistance = rstretch_count * 2 + 2 - (rstretch_count - (poff_stretch_count - 1) );
							p_poffset = rstretch_count - (rstretch_count - poff_stretch_count ) * 2;
						} else {
							p_rdistance = rstretch_count * 2 + 1 - (rstretch_count - poff_stretch_count);
							p_poffset = rstretch_count - (rstretch_count - poff_stretch_count ) * 2;
						}
					}
					break;
				case 2:
						p_rdistance = rstretch_count * 2 + 1 - (rstretch_count - poff_stretch_count);
						p_poffset = rstretch_count - (rstretch_count - poff_stretch_count) * 2;
					break;
				}
			}

			break;

		case 2:
			if (
				( rstretch_count % 2 == 0 &&
					(
						(poff_stretch_count < rstretch_count / 2) ||
						(poff_stretch_count == rstretch_count / 2 &&
							poff_stretch < 2 )
					)
				) ||
				( rstretch_count % 2 == 1 &&
					(
						(poff_stretch_count <= rstretch_count / 2 ) ||
						(poff_stretch_count == rstretch_count / 2 + 1 &&
							poff_stretch == 0 && poffoff < roff )
					)
				)
				)
			{
				switch (poff_stretch) {
				case 0:
					if (poffoff < roff) {
						p_rdistance = (rstretch_count + 1) * 2 - poff_stretch_count;
						p_poffset = rstretch_count + poff_stretch_count;
					} else {
						p_rdistance = (rstretch_count + 1) * 2 - poff_stretch_count - 1;
						p_poffset = rstretch_count + poff_stretch_count;
					}
					break;

				case 1:
					p_rdistance = (rstretch_count + 1) * 2 - poff_stretch_count - 1;
					p_poffset = rstretch_count + poff_stretch_count;
					break;

				case 2:
					p_rdistance = (rstretch_count + 1) * 2 - poff_stretch_count - 1;
					p_poffset = rstretch_count + poff_stretch_count + 1;
					break;
				}
			} else {
				p_phase = phase < 5 ? phase + 1 : 0;

				switch (poff_stretch) {
				case 0:
					if (poffoff < roff) {
						p_rdistance = (rstretch_count + 1) * 2 - (rstretch_count - (poff_stretch_count - 1));
						p_poffset = rstretch_count - (rstretch_count - (poff_stretch_count - 1)) * 2;
					} else {
						p_rdistance = (rstretch_count + 1) * 2 - (rstretch_count - (poff_stretch_count - 1));
						p_poffset = rstretch_count - (rstretch_count - (poff_stretch_count - 1)) * 2 + 1;
					}
					break;

				case 1:
					p_rdistance = (rstretch_count + 1) * 2 - (rstretch_count - (poff_stretch_count - 1));
					p_poffset = rstretch_count - (rstretch_count - (poff_stretch_count - 1)) * 2 + 1;
					break;

				case 2:
					p_rdistance = (rstretch_count + 1) * 2 - (rstretch_count - (poff_stretch_count - 1)) + 1;
					p_poffset = rstretch_count - (rstretch_count - poff_stretch_count) * 2;
					break;
				}
			}
			break;
		}

		} else { // direction reversed...

		switch (rstretch) {
		case 0:
			if (
				( rstretch_count % 2 == 1 && (
					(poff_stretch_count < rstretch_count / 2) ||
					(poff_stretch_count == rstretch_count / 2 &&
					(poff_stretch < 2 || poffoff < roff) )
				)) ||
				( rstretch_count % 2 == 0 &&
					(
						(poff_stretch_count < rstretch_count / 2) ||
						(poff_stretch_count == rstretch_count / 2 &&
							poff_stretch == 0
						)
					)
				)
			)
			{
				p_phase = phase > 0 ? phase - 1 : 5;

				switch (poff_stretch) {
				case 0:
					p_rdistance = rstretch_count * 2 - poff_stretch_count;
					p_poffset = rstretch_count + poff_stretch_count - 1;
					break;
				case 1:
					p_rdistance = rstretch_count * 2 - poff_stretch_count;
					p_poffset = rstretch_count + poff_stretch_count;

					break;
				case 2:
					if (poffoff < roff) {
						p_rdistance = rstretch_count * 2 - poff_stretch_count;
						p_poffset = rstretch_count + poff_stretch_count;
					} else {
						p_rdistance = rstretch_count * 2 - poff_stretch_count - 1;
						p_poffset = rstretch_count + poff_stretch_count;
					}
					break;
				}
			}
			else
			{
				switch (poff_stretch) {
				case 0:
					p_rdistance = rstretch_count + poff_stretch_count;
					p_poffset = poff_stretch_count * 2 - rstretch_count - 1;
					break;
				case 1:
					p_rdistance = rstretch_count + poff_stretch_count + 1;
					p_poffset = poff_stretch_count * 2 - rstretch_count;
					break;
				case 2:
					if (poffoff < roff) {
						p_rdistance = rstretch_count + poff_stretch_count + 1;
						p_poffset = poff_stretch_count * 2 - rstretch_count;
					} else {
						p_rdistance = rstretch_count + poff_stretch_count + 1;
						p_poffset =  poff_stretch_count * 2 - rstretch_count + 1;
					}
					break;
				}
			}
			break;

		case 1:
			if (
				(rstretch_count % 2 == 0 &&
					(
						(poff_stretch_count < rstretch_count / 2) ||
						(poff_stretch_count == rstretch_count / 2 &&
							(poff_stretch == 0 ||
									(poff_stretch == 1 && poffoff < roff )
							)
						)
					)
				) ||
				(rstretch_count % 2 == 1 &&
					poff_stretch_count <= rstretch_count / 2

				)
			)
			{
				p_phase = phase > 0 ? phase - 1 : 5;

				switch (poff_stretch) {
				case 0:
					p_rdistance = 1 + rstretch_count * 2 - poff_stretch_count;
					p_poffset = rstretch_count + poff_stretch_count;
					break;
				case 1:
					if (poffoff < roff) {
						p_rdistance = rstretch_count * 2 - poff_stretch_count + 1;
						p_poffset = rstretch_count + poff_stretch_count;
					} else {
						p_rdistance = rstretch_count * 2 - poff_stretch_count;
						p_poffset = rstretch_count + poff_stretch_count;
					}
					break;
				case 2:
					p_rdistance = rstretch_count * 2 - poff_stretch_count;
					p_poffset = rstretch_count + poff_stretch_count;
					break;
				}
			}
			else
			{
				switch (poff_stretch) {
				case 0:
					p_rdistance = rstretch_count + poff_stretch_count + 1;
					p_poffset = poff_stretch_count * 2 - rstretch_count - 1;
					break;
				case 1:
					if (poffoff < roff) {
						p_rdistance = rstretch_count + poff_stretch_count + 1;
						p_poffset =  poff_stretch_count * 2 - rstretch_count - 1;
					} else {
						p_rdistance = rstretch_count + poff_stretch_count + 1;
						p_poffset =  poff_stretch_count * 2 - rstretch_count;
					}
					break;
				case 2:
					p_rdistance = rstretch_count + poff_stretch_count + 1;
					p_poffset = poff_stretch_count * 2 - rstretch_count;
					break;
				}

			}

			break;
		case 2:
			if (
				( rstretch_count % 2 == 0 &&
					(poff_stretch_count < rstretch_count / 2 ||
						(poff_stretch_count == rstretch_count / 2 && poff_stretch < 2)
					)
				) ||
				( rstretch_count % 2 == 1 &&
					(
						(poff_stretch_count <= rstretch_count / 2) ||
						(poff_stretch_count == rstretch_count / 2 + 1 &&
						(poff_stretch == 0 && poffoff < roff) )
					)
				)
			)
			{
				p_phase = phase > 0 ? phase - 1 : 5;
				switch (poff_stretch) {
				case 0:
					if (poffoff < roff) {
						p_rdistance = 2 + rstretch_count * 2 - poff_stretch_count;
						p_poffset = rstretch_count + poff_stretch_count;
					} else {
						p_rdistance = 1 + rstretch_count * 2 - poff_stretch_count;
						p_poffset = rstretch_count + poff_stretch_count;
					}
					break;
				case 1:
					p_rdistance = 1 + rstretch_count * 2 - poff_stretch_count;
					p_poffset = rstretch_count + poff_stretch_count;
					break;
				case 2:
					p_rdistance = rstretch_count * 2 - poff_stretch_count + 1;
					p_poffset = rstretch_count + poff_stretch_count + 1;
					break;
				}
			}
			else
			{
				switch (poff_stretch) {
				case 0:
					if (poffoff < roff) {
						p_rdistance = rstretch_count + poff_stretch_count + 1;
						p_poffset = poff_stretch_count * 2 - rstretch_count - 2;
					} else {
						p_rdistance = rstretch_count + poff_stretch_count + 1;
						p_poffset = poff_stretch_count * 2 - rstretch_count - 1;
					}
					break;
				case 1:
					p_rdistance = rstretch_count + poff_stretch_count + 1;
					p_poffset = poff_stretch_count * 2 - rstretch_count - 1;
					break;
				case 2:
					p_rdistance = rstretch_count + poff_stretch_count + 2;
					p_poffset = poff_stretch_count * 2 - rstretch_count;
					break;
				}
			}
			break;
		}

		}
	}

	return HexSpiralCoordinate(p_rdistance, p_phase, p_poffset);
}

HexSpiralCoordinate HexSpiralCoordinate::getChildSpiralCoordinate(bool direction_forward, int map_radius)
{
	int c_rdistance = 0, c_phase = phase, c_poffset = 0;

	if (!direction_forward) {
		if (rdistance == 1) {
			c_rdistance = map_radius * 2;
			c_poffset = map_radius - 1;
		} else if (rdistance > 1) {
			if (poffset < rdistance / 2) { // phase shift down?
				c_phase = c_phase > 0 ? c_phase - 1 : 5;
				c_rdistance = map_radius * ( (rdistance * 2 - 1) - poffset ); // ???
				c_poffset = map_radius * ( (rdistance + 1) + poffset) - 1;
			} else { // no phase change
				c_rdistance = map_radius * (rdistance * 2 - (rdistance - 1 - poffset)  );
				c_poffset = map_radius * (rdistance - ( rdistance - 1 - poffset ) * 2 ) - 1;
			}
		}
	} else {
		if (rdistance == 1) {
			c_phase = phase < 5 ? phase + 1 : 0;
			c_rdistance = map_radius * 2;
			c_poffset = map_radius - 1;
		} else if (rdistance > 1) {
			if (poffset <= (rdistance - 2) / 2 ) {
				// phase unchanged
				c_rdistance = map_radius * ( (rdistance - 1) * 2 + 1 - poffset );
				c_poffset = map_radius * ( rdistance + 1 + poffset ) - 1;
			} else {
				c_phase = phase < 5 ? phase + 1 : 0;
				c_rdistance = map_radius * (  (rdistance * 2) - ( rdistance - poffset - 1 ) );
				c_poffset = map_radius * (rdistance - ( rdistance - poffset - 1 ) * 2  ) - 1;
			}
		}
	}

	return HexSpiralCoordinate(c_rdistance, c_phase, c_poffset);
}


HexSpiralCoordinate HexSpiralCoordinate::findWrappedPosition(int radius, bool invert) {

	HexSpiralCoordinate result = (
		(*this)
			-
		(*this)
			.getParentSpiralCoordinate(!invert, radius)
			.getChildSpiralCoordinate(invert, radius)
	);

	if (result.rdistance == radius && result.phase < 3) {
		int realphase = result.phase + 3;
		int realpoffset = result.rdistance - 2 - result.poffset;
		if (realpoffset < 0) {
			realphase--;
			realpoffset = result.rdistance - 1;
		}

		return HexSpiralCoordinate(result.rdistance, realphase, realpoffset);
	} else
		return result;
}
