#include "RealmMapDisplay.h"
#include "RealmSceneView.h"
#include "HexGridDisplay.h"
#include "RealmSceneElevationMap.h"
#include "RealmColorMap.h"
#include "RealmSceneBackground.h"
#include "RealmBackgroundScroller.h"
#include "RealmView.h"

RealmMapDisplay::RealmMapDisplay(
							RealmSceneView* scene_view,
							RealmView* master_view,
							RealmSceneElevationMapRepository* maprepo,
							int base_radius,
							int grid_size, bool inverted, int maps_radius,
							RealmSceneKey scene_center
						) :
	SceneCenter(scene_center)
{
	SceneView = scene_view;
	MasterView = master_view;
	MapRepository = maprepo;
	BackgroundRepository = MapRepository->getBackgroundRepository();
	BackgroundScroller = new RealmBackgroundScroller(base_radius, grid_size, inverted, maps_radius);
	BackgroundScroller->setMapDisplay(this);
	Active = false;
}

RealmMapDisplay::~RealmMapDisplay() {
	delete BackgroundScroller;
}

void RealmMapDisplay::updateSceneCenter() {
	RealmLocation current_loc = MasterView->getCurrentLocation();
	RealmMapKey mapkey(current_loc, SceneView->getViewMode());

	if (CurrentMap)
		mapkey = CurrentMap->getMapKey();

	SceneCenter = mapkey.MapCenterKey;
}

double RealmMapDisplay::getZoomFactor() {
	return SceneView->getZoomScale(SceneView->getBaseLevel());
}

void RealmMapDisplay::activate() {
	Active = true;
	updateSceneCenter();
}

void RealmMapDisplay::toggle() {
	Active = !Active;
	if (Active)
		updateSceneCenter();
}

void RealmMapDisplay::disable() {
	Active = false;
}

void RealmMapDisplay::update() {
	if (Active && BackgroundScroller)
		BackgroundScroller->update(SceneView);
}

void RealmMapDisplay::updateBaseRadius(int base_radius) {
	BackgroundScroller->setHexRadius(base_radius);
}

void RealmMapDisplay::reCenterCurrentMap() {
	RealmLocation current_loc = MasterView->getCurrentLocation();
	RealmMapKey mapkey(current_loc, SceneView->getViewMode());

	mapkey.CenterSpiralOffset =
		current_loc
			.getHexSpiralCoordinate(SceneView->getViewMode())
				.getParentSpiralCoordinate(!RealmLocation::isInvertedGrid(SceneView->getViewMode()))
					.getHexIndex();

	mapkey.MapCenterKey.setSpiralIndex(current_loc.getHexSpiralCoordinate(SceneView->getViewMode()+1).getHexIndex());

	SceneCenter = mapkey.MapCenterKey;
	CurrentMap = MapRepository->findSceneElevationMap(mapkey);
	BackgroundScroller->setCenter(mapkey.CenterSpiralOffset, true);
}

void RealmMapDisplay::refreshCurrentMap() {
	RealmLocation current_loc = MasterView->getCurrentLocation();
	RealmMapKey mapkey(current_loc, SceneView->getViewMode());

	if (CurrentMap) {
		mapkey = CurrentMap->getMapKey();
		mapkey.CenterSpiralOffset =
			current_loc
				.getHexSpiralCoordinate(SceneView->getViewMode())
					.getParentSpiralCoordinate(!RealmLocation::isInvertedGrid(SceneView->getViewMode()))
						.getHexIndex();
	}

	SceneCenter = mapkey.MapCenterKey;
	CurrentMap = MapRepository->findSceneElevationMap(mapkey);
	BackgroundScroller->setCenter(mapkey.CenterSpiralOffset, true); // trigger redraw
}



void RealmMapDisplay::populateStep() {
	if (Active) {
		refreshCurrentMap();

		CurrentMap->populateStep();
	}
}

void RealmMapDisplay::populateToStep(int level) {
	if (CurrentMap)
		CurrentMap->populateToStep(level);
}

void RealmMapDisplay::populateNeighborStep() {
	refreshCurrentMap();

	if (CurrentMap)
		CurrentMap->populateNeighborStep();
}


RealmSceneBackground* RealmMapDisplay::getSceneBackground(long int idx)
{
	RealmMapKey mapkey(SceneCenter, idx);
	return BackgroundRepository->findSceneBackground(mapkey);
}

void RealmMapDisplay::updateTextures(SDL_Renderer* renderer) {
	if (Active && BackgroundScroller)
		BackgroundScroller->updateTextures(renderer);
}

void RealmMapDisplay::draw(SDL_Renderer* renderer) {
	if (Active && BackgroundScroller)
		BackgroundScroller->draw(renderer);
}
