#include "XYAnimator.h"
#include <math.h>

XYAnimator::XYAnimator(): LastLocation(0, 0) {
	MovementSpeed = 0;
	XOffset = 0; YOffset = 0;
	MaxMovementSpeed = 20;
}


void XYAnimator::doMove(XYLocation target)
{
	TargetPixel = target;
	//MovementSpeed++;
	MovementSpeed = 1;
	LastLocation.x = LastLocation.x + XOffset;
	LastLocation.y = LastLocation.y + YOffset;
	XOffset = YOffset = 0;

	int xdistance = target.x - LastLocation.x;
	int ydistance = target.y - LastLocation.y;

	if (abs(xdistance) > 0 || abs(ydistance) > 0) {
		double yxratio;
		double angle = 0;
		if (abs(xdistance) > 0) {
			yxratio = (double) abs(ydistance) / (double) abs(xdistance);
			angle = atan(yxratio);
		} else {
			if (abs(ydistance) > 0)
				angle = atan(1) * 2;
		}

		XMoveRatio = cos(angle) * 3;
		YMoveRatio = sin(angle) * 3;
	}
}

bool XYAnimator::advanceFrame()
{
	XYLocation CurrentLocation(LastLocation.x + XOffset, LastLocation.y + YOffset);

	int xdistance = TargetPixel.getx() - CurrentLocation.x;
	int ydistance = TargetPixel.gety() - CurrentLocation.y;

	if (abs(xdistance) > 0 || abs(ydistance) > 0) {
		double xmove = MovementSpeed * XMoveRatio;
		double ymove = MovementSpeed * YMoveRatio;

		if (CurrentLocation.x < TargetPixel.getx()) {
			if (CurrentLocation.x + xmove > TargetPixel.getx())
				xmove = TargetPixel.getx() - CurrentLocation.x;
			XOffset += xmove;
		} else if (CurrentLocation.x > TargetPixel.getx()) {
			if (CurrentLocation.x - xmove < TargetPixel.getx())
				xmove = CurrentLocation.x - TargetPixel.getx();
			XOffset -= xmove;
		}

		if (CurrentLocation.y < TargetPixel.gety()) {
			if (CurrentLocation.y + ymove > TargetPixel.gety())
				ymove = TargetPixel.gety() - CurrentLocation.y;
			YOffset += ymove;
		} else if (CurrentLocation.y > TargetPixel.gety()) {
			if (CurrentLocation.y - ymove < TargetPixel.gety())
				ymove = CurrentLocation.y - TargetPixel.gety();
			YOffset -= ymove;
		}
	}

	if ( (abs(xdistance) > 50 || abs(ydistance) > 50 ) && MovementSpeed < MaxMovementSpeed) {
		MovementSpeed++;
	} else if ((abs(xdistance) < 30 && abs(ydistance) < 30 ) && MovementSpeed > 2) {
		MovementSpeed--;
	}

	// if we have reached our target, reset variables to stop animating
	if (abs(xdistance) == 0 && abs(ydistance) == 0 ) {
		MovementSpeed = 0;
		LastLocation = TargetPixel;
		XOffset = 0;
		YOffset = 0;
		return true;
	}

	return false;
}

